/* 
 * File:   CopyFiles.cpp
 * Author: Saleem EDAH-TALLY - nmset@yandex.com
 * License : LGPL
 * Copyright Saleem EDAH-TALLY - © 2014
 * 
 * Created on 28 septembre 2014, 17:53
 */

#include "CopyFiles.h"
#include <wx/wfstream.h>
#include <wx/filename.h>
#include <iostream>
using namespace std;

FileCopy::FileCopy(wxWindow * parent, FileCopyEvent * evh,
                   const wxString& src, const wxString& dest,
                   const wxPoint progDialogPosition,
                   bool move, const int buffSize)
: wxThread(wxTHREAD_DETACHED)
{
    m_src = src;
    m_dest = dest;
    m_move = move;
    m_bufSize = buffSize;
    m_owner = parent;
    m_evh = evh;
    wxFileName srcFn(m_src);

    m_pdlg = new wxProgressDialog(m_src, wxEmptyString, 100000, NULL, wxPD_AUTO_HIDE | wxPD_CAN_ABORT | wxPD_ESTIMATED_TIME | wxPD_REMAINING_TIME);
    m_pdlg->SetPosition(progDialogPosition);
    m_pdlg->SetTitle(srcFn.GetFullName());
    m_fcp = new FileCopyProgress(m_pdlg, srcFn.GetSize().ToULong());
}

FileCopy::~FileCopy()
{
    delete m_fcp;
    if (m_evh) m_evh->OnDeath(GetId());
    delete m_evh;
    //cout << "DTOR : FileCopy" << endl;
}

/*
 * I won't swear this code is error free.
 * But it has not failed me since 2015.
 */
wxThread::ExitCode FileCopy::Entry()
{
    if (m_evh) m_evh->OnRun(GetId(), m_pdlg); // Show pdlg there
    wxFileName destFileName(m_dest);
    wxFileName srcFileName(m_src);
    if (destFileName.Exists())
    {
        if (m_evh) m_evh->OnAbortExisting(m_src, m_dest);
        return (wxThread::ExitCode) 0;
    }
    if (!srcFileName.Exists())
    {
        if (m_evh) m_evh->OnError(m_src, m_dest);
        return (wxThread::ExitCode) 0;
    }
    wxFFileInputStream instr(m_src);
    wxFFileOutputStream outstr(m_dest);
    char * buf = new char[m_bufSize];
    unsigned long done = 0;
    if (instr.IsOk() && outstr.IsOk())
    {
        while (true)
        {
            if (TestDestroy()) break;
            instr.Read(buf, m_bufSize);
            if (instr.GetLastError() == wxSTREAM_READ_ERROR)
            {
                outstr.Close();
                if (m_evh) m_evh->OnError(m_src, m_dest);
                break;
            }
            const size_t nb = instr.LastRead();
            if (TestDestroy()) break;
            outstr.Write(buf, nb);
            if (outstr.GetLastError() == wxSTREAM_WRITE_ERROR)
            {
                if (m_evh) m_evh->OnError(m_src, m_dest);
                break;
            }
            done += nb;
            m_fcp->SetParams(done);
            m_fcp->CallAfter(&FileCopyProgress::DoUpdate);

            if (TestDestroy()) break;
            if (instr.Eof() || instr.GetLastError() == wxSTREAM_EOF)
            {
                if (m_move)
                {
                    outstr.Sync();
                    if (wxFileName::Exists(m_src)) wxRemoveFile(m_src);
                }
                if (m_evh) m_evh->OnSuccess(m_src, m_dest);
                break;
            }
            if (m_pdlg->WasCancelled())
            {
                if (m_evh) m_evh->OnCancel(m_src, m_dest);
                if (!outstr.Close())
                {
                    if (m_evh) m_evh->OnError(m_src, m_dest);
                    break;
                }
                break;
            }
        }
    }
    else
    {
        outstr.Close();
        if (m_evh) m_evh->OnError(m_src, m_dest);
    }

    delete buf;
    return (wxThread::ExitCode) 0; // Only relevant for joinable threads
}

FileCopyEvent::FileCopyEvent()
{

}

FileCopyEvent::~FileCopyEvent()
{

}

FileCopyProgress::FileCopyProgress(wxProgressDialog* pdlg, const unsigned long fileSize)
{
    m_pdlg = pdlg;
    m_done = 0;
    m_fileSize = fileSize;
}

FileCopyProgress::~FileCopyProgress()
{
    if (m_pdlg)
    {
        m_pdlg->Show(false);
        m_pdlg->Destroy();
    }
}

void FileCopyProgress::SetParams(unsigned long done)
{
    m_done = done;
}

void FileCopyProgress::DoUpdate()
{
    if (m_pdlg)
    {
        const uint idone = ((double) m_done / (double) m_fileSize) * 100000;
        m_pdlg->Update(idone, wxFileName::GetHumanReadableSize((wxULongLong) m_done));
    }

}

ProgressDialogTiles::ProgressDialogTiles()
{
    m_current = 0;
    wxProgressDialog pdlg(wxEmptyString, wxEmptyString, 100,
                          NULL, wxPD_AUTO_HIDE | wxPD_CAN_ABORT | wxPD_ESTIMATED_TIME | wxPD_REMAINING_TIME);
    pdlg.Show(false);
    m_defSize = pdlg.GetSize();
    const wxSize screenSize = wxGetDisplaySize();
    m_hMax = screenSize.GetWidth() / m_defSize.GetWidth();
    m_vMax = ((double) screenSize.GetHeight() * 1.1) / m_defSize.GetHeight();
    const int hResidual = screenSize.GetWidth() - (m_hMax * m_defSize.GetWidth());
    for (unsigned int h = m_hMax; h > 0; h--)
    {
        for (unsigned int v = 0; v < m_vMax; v++)
        {
            wxPoint origin(((h - 1) * m_defSize.GetWidth()) + hResidual, v * m_defSize.GetHeight());
            wxRect tile(origin, m_defSize);
            m_tiles.push_back(tile);
        }
    }
}

ProgressDialogTiles::~ProgressDialogTiles()
{
    //cout << "DTOR : ProgressDialogTiles" << endl;
}

const wxRect ProgressDialogTiles::GetCurrentTile()
{
    const wxRect tile = m_tiles.at(m_current);
    m_current++;
    if (m_current == m_tiles.size()) m_current = 0;
    return tile;
}
