/* 
 * File:   FileWorker.cpp
 * Author: Saleem EDAH-TALLY - nmset@yandex.com
 * License : GPL
 * Copyright Saleem EDAH-TALLY - © 2017
 * 
 * Created on 5 mars 2017, 20:56
 */

#include "gcc_version.h"
#include "FileWorker.h"
#include "globals.h"
#include "XCB7.h"
#include "MiscTools.h"
#include "ConfigEditorPopup.h"
#include <wx/dir.h>
#include <wx/stdpaths.h>
#include <iostream>
using namespace std;

FileWorker::FileWorker(XCB7 * owner)
{
    m_owner = owner;
    m_szMain = m_owner->szMain;
    m_paneFiles = m_owner->paneFiles;
    m_cbMove = m_owner->cbMove;
    m_cbTimeStamp = m_owner->cbTimeStamp;
    m_cbAppendName = m_owner->cbAppendName;
    m_cbPreferOriginal = m_owner->cbPreferOriginal;
    m_txtResults = m_owner->txtResults;
    m_btnSelectFile = m_owner->btnSelectFile;
    m_config = m_owner->GetConfig();
    m_tiles = NULL;

    m_paneFiles->Bind(wxEVT_COLLAPSIBLEPANE_CHANGED, &FileWorker::OnPaneFilesChanged, this);
    m_paneFiles->SetDropTarget(new DNDFileDataTransfer(this, m_paneFiles));

    m_cbMove->SetValue(m_config->ReadBool(_T("/Files/Move"), 0));
    m_cbTimeStamp->SetValue(m_config->ReadBool(_T("/Files/TimeStamp"), 1));
    m_cbAppendName->SetValue(m_config->ReadBool(_T("/Files/AppendOriginalName"), 0));
    m_cbPreferOriginal->SetValue(m_config->ReadBool(_T("/Files/NoRename"), 0));

    m_cbMove->Bind(wxEVT_CHECKBOX, &FileWorker::OnCbMove, this);
    m_cbTimeStamp->Bind(wxEVT_CHECKBOX, &FileWorker::OnCbTimeStamp, this);
    m_cbAppendName->Bind(wxEVT_CHECKBOX, &FileWorker::OnCbAppendName, this);
    m_cbPreferOriginal->Bind(wxEVT_CHECKBOX, &FileWorker::OnCbPreferOriginal, this);

    m_txtResults->Show(false);

    m_btnSelectFile->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &FileWorker::OnBtnSelectFile, this);
}

FileWorker::~FileWorker()
{
    delete m_tiles;
}

bool FileWorker::DoRenameAndCopy(const wxArrayString& filenames)
{
    if (m_owner->cmbFolder->GetValue().IsEmpty())
    {
        MiscTools::MessageBox(_("Missing folder name."), true);
        return false;
    }
    if (!m_cbPreferOriginal->IsChecked()
        && m_owner->cmbLabel->GetValue().IsEmpty())
    {
        MiscTools::MessageBox(_("Missing label."), true);
        return false;
    }
    wxString destDir = m_owner->dpkRepository->GetPath() + _SEP_
            + wxVariant(m_owner->spnYear->GetValue()).GetString() + _SEP_
            + m_owner->cmbFolder->GetValue();
    if (!wxDir::Exists(destDir)
        &&!wxFileName::Mkdir(destDir, wxS_DIR_DEFAULT, wxPATH_MKDIR_FULL))
        return false;
    for (uint i = 0; i < filenames.GetCount(); i++)
    {
        /*
         * m_tiles may be deleted in FileCopyEVH::OnDeath
         * This happens when copying multiple tiny files.
         * m_tiles gets destroyed before the loop completes -> crash
         */
        if (!m_tiles)
            m_tiles = new ProgressDialogTiles();
        wxFileName src(filenames.Item(i));
        wxString destFullname;
        if (!m_cbPreferOriginal->IsChecked())
        {
            wxString destBasename = m_owner->cmbLabel->GetValue();
            if (m_cbTimeStamp->IsChecked())
                destBasename += _DASH_ + wxVariant(src.GetModificationTime().GetTicks()).GetString();
            if (!m_owner->cmbHint->GetValue().IsEmpty())
                destBasename += _DASH_ + m_owner->cmbHint->GetValue();
            if (m_cbAppendName->IsChecked())
            {
                wxString srcBasename = MiscTools::GetRealBaseName(src);
                srcBasename.Replace(_SPACE_, _UNDERSCORE_, true);
                destBasename += _DASH_ + srcBasename;
            }
            destFullname = destBasename;
            if (src.HasExt())
            {
                destFullname += _DOT_ + MiscTools::GetFullExt(src);
            }
        }
        else
        {
            destFullname = src.GetFullName();
        }
        const wxString destFullPath = destDir + _SEP_ + destFullname;
        const int buffSize = src.GetSize() > (3 * 1024 * 1024)
                ? (1024 * 1024)
                : (50 * 1024);
        FileCopy * fc = new FileCopy(m_owner, new FileCopyEVH(this, m_owner),
                                     filenames.Item(i),
                                     destFullPath, m_tiles->GetCurrentTile().GetPosition(),
                                     m_cbMove->IsChecked(), buffSize);
        fc->Run();
    }
    return true;
}

void FileWorker::OnPaneFilesChanged(wxCollapsiblePaneEvent& evt)
{
    m_txtResults->Show(m_paneFiles->IsExpanded());
    m_owner->szMain->Layout();
    evt.Skip();
}

void FileWorker::OnCbMove(wxCommandEvent& evt)
{
    m_config->Write(_T("/Files/Move"), evt.IsChecked());
    m_config->Flush();
    evt.Skip();
}

void FileWorker::OnCbTimeStamp(wxCommandEvent& evt)
{
    m_config->Write(_T("/Files/TimeStamp"), evt.IsChecked());
    if (evt.IsChecked())
    {
        m_cbPreferOriginal->SetValue(false);
        m_config->Write(_T("/Files/NoRename"), false);
    }
    evt.Skip();
}

void FileWorker::OnCbAppendName(wxCommandEvent& evt)
{
    m_config->Write(_T("/Files/AppendOriginalName"), evt.IsChecked());
    if (evt.IsChecked())
    {
        m_cbPreferOriginal->SetValue(false);
        m_config->Write(_T("/Files/NoRename"), false);
    }
    m_config->Flush();
    evt.Skip();
}

void FileWorker::OnCbPreferOriginal(wxCommandEvent& evt)
{
    m_config->Write(_T("/Files/NoRename"), evt.IsChecked());
    if (evt.IsChecked())
    {
        m_cbAppendName->SetValue(false);
        m_cbTimeStamp->SetValue(false);
        m_config->Write(_T("/Files/AppendOriginalName"), false);
        m_config->Write(_T("/Files/TimeStamp"), false);
    }
    m_config->Flush();
    evt.Skip();
}

void FileWorker::OnBtnSelectFile(wxCommandEvent& evt)
{
    const wxString last = m_config->Read(_T("/Last/Files"), wxEmptyString);
    wxFileDialog fd(m_owner, _("Select files :"),
                    wxEmptyString, last.IsEmpty()
                    ? wxStandardPaths::Get().GetUserDataDir()
                    : last + wxFileName::GetPathSeparator(),
                    wxEmptyString, wxFD_MULTIPLE | wxFD_PREVIEW | wxFD_FILE_MUST_EXIST);
    if (fd.ShowModal() != wxID_OK)
    {
        evt.Skip();
        return;
    }
    m_config->Write(_T("/Last/Files"), fd.GetDirectory());
    m_config->Flush();
    wxArrayString src;
    fd.GetPaths(src);
    DoRenameAndCopy(src);
    evt.Skip();
}


////////////////////////////////////////////////////////////////////////////////

FileCopyEVH::FileCopyEVH(FileWorker * caller, wxWindow * owner)
{
    m_caller = caller;
    m_owner = owner;
}

FileCopyEVH::~FileCopyEVH()
{

}

void FileCopyEVH::OnRun(wxThreadIdType id, wxProgressDialog * pdlg)
{
    m_caller->m_threads.Add(id);
    pdlg->Show(true);
}

void FileCopyEVH::OnDeath(wxThreadIdType id)
{
    m_caller->m_threads.Remove(id);
    if (m_caller->m_threads.Count() == 0) wxDELETE(m_caller->m_tiles);
}

void FileCopyEVH::OnError(const wxString& src, const wxString& dest)
{
    const wxString msg = _("Error : ") + src + _NL_ + _T(" -> ") + dest + _NL_;
    XCB7 * app = static_cast<XCB7*> (m_owner);
    app->CallAfter(&XCB7::ForwardFileCopyInfo, msg);
}

void FileCopyEVH::OnCancel(const wxString& src, const wxString& dest)
{
    if (wxFileName::Exists(src) && (wxFileName::Exists(dest))) wxRemoveFile(dest);
    const wxString msg = _("Abort : ") + src + _NL_ + _T(" -> ") + dest + _NL_;
    XCB7 * app = static_cast<XCB7*> (m_owner);
    app->CallAfter(&XCB7::ForwardFileCopyInfo, msg);
}

void FileCopyEVH::OnSuccess(const wxString& src, const wxString& dest)
{
    const wxString msg = _("Success : ") + src + _NL_ + _T(" -> ") + dest + _NL_;
    XCB7 * app = static_cast<XCB7*> (m_owner);
    app->CallAfter(&XCB7::ForwardFileCopyInfo, msg);
    wxFileName destFileName(dest);
    destFileName.SetPermissions(wxPOSIX_USER_READ | wxPOSIX_USER_WRITE | wxPOSIX_GROUP_READ | wxPOSIX_OTHERS_READ);
}

void FileCopyEVH::OnAbortExisting(const wxString& src, const wxString& dest)
{
    const wxString msg = _("Existing : ") + src + _NL_ + _T(" -> ") + dest + _NL_;
    XCB7 * app = static_cast<XCB7*> (m_owner);
    app->CallAfter(&XCB7::ForwardFileCopyInfo, msg);
}

////////////////////////////////////////////////////////////////////////////////

DNDFileDataTransfer::DNDFileDataTransfer(FileWorker * parent, wxWindow * source)
{
    m_owner = parent;
    m_pane = source;
}

DNDFileDataTransfer::~DNDFileDataTransfer()
{
}

bool DNDFileDataTransfer::OnDropFiles(wxCoord x, wxCoord y, const wxArrayString& filenames)
{
    m_owner->m_txtResults->Show(true);
    m_owner->m_szMain->Layout();
    return m_owner->DoRenameAndCopy(filenames);
}
