/*
* File:   InsaneWorker.cpp
* Author: Saleem Edah-Tally - nmset@yandex.com
* License : CeCILL-C
* Copyright Saleem Edah-Tally - © 2023
*
* Created on 14 02 2024, 13:00
*/

#include "InsaneWorker.h"
#include <iostream>
#include <libinsane/log.h>
#include <libinsane/safebet.h>
#include <libinsane/error.h>
#include <libinsane/util.h>
#include <libinsane/normalizers.h>
#include <libinsane/constants.h>

using namespace std;

#define IERR(e) ("[" + to_string(e) + "] " + lis_strerror(e))

InsaneWorker::InsaneWorker(const std::string& deviceId,
                           const std::string& filePath, OUTPUT_FILETYPE fileType,
                           const std::string& source,
                           const std::string& mode, int resolution,
                           InsaneWorkerEvent * evh, lis_api * api)
{
    m_api = api;
    m_evh = evh;
    m_deviceId = deviceId;
    m_fileName = filePath;
    m_fileType = fileType;
    m_source = source;
    m_mode = mode;
    m_resolution = resolution;
}

InsaneWorker::~InsaneWorker()
{

}

bool InsaneWorker::Init()
{
    // Clean existing api.
    if (m_api)
    {
        m_api->cleanup(m_api);
        m_api = nullptr;
    }
    lis_error err;
    
    // Create new api;
    err = lis_safebet(&m_api);
    if (err)
    {
        if (m_evh)
            m_evh->OnInsaneError(IERR(err));
        return false;
    }
    
    err = m_api->get_device(m_api, m_deviceId.c_str(), &m_item);
    if (err)
    {
        if (m_evh)
            m_evh->OnInsaneError(IERR(err));
        return false;
    }
    
    // We are dealing with single byte characters.
    auto tolower = [] (const string& input)
    {
        string output;
        for (char c : input)
            output.append(std::tolower(c), 1);
        return output;
    };
    
    // Load options.
    lis_option_descriptor ** options;
    uint numberOfOptions = 0;
    auto loadOptions = [&] ()
        {
        err = m_item->get_options(m_item, &options);
        if (!err)
        {
            for (uint i = 0; options[i] != nullptr; i++)
            {
                numberOfOptions++;
            }
        }
    };
    loadOptions();
    if (err)
    {
        FREE(options);
        if (m_evh)
            m_evh->OnInsaneError(IERR(err));
        return false;
    }
    
    // Set source, mode and resolution if different from defaults.
    bool mustReloadOptions = false;
    for (uint i = 0; i < numberOfOptions; i++)
    {
        lis_option_descriptor * optionDescriptor = options[i];
        if (!LIS_OPT_IS_WRITABLE(optionDescriptor))
            continue;
        const string name = optionDescriptor->name;
        if (tolower(name) != string(OPT_NAME_SOURCE)
            && tolower(name) != string(OPT_NAME_MODE)
            && tolower(name) != string(OPT_NAME_RESOLUTION))
            continue;
        
        union lis_value value;
        err = optionDescriptor->fn.get_value(optionDescriptor, &value);
        if (err)
        {
            FREE(options);
            if (m_evh)
                m_evh->OnInsaneError(IERR(err));
            return false;
        }
        
        if (tolower(name) == string(OPT_NAME_SOURCE))
        {
            int setFlags = 0;
            union lis_value newValue;
            newValue.string = m_source.c_str();
            
            if (!LIS_OPT_IS_READABLE(optionDescriptor))
            {
                const string currentValue = value.string;
                if (currentValue != m_source)
                {
                    err = optionDescriptor->fn.set_value(optionDescriptor, newValue, &setFlags);
                }
            }
            else
            {
                err = optionDescriptor->fn.set_value(optionDescriptor, newValue, &setFlags);
            }
            if (err) // LIS_APPROXIMATE_VALUE, LIS_ACCESS_DENIED, LIS_INVALID_VALUE
            {
                // Don't free options nor return here; use default values.
                if (m_evh)
                    m_evh->OnInsaneError(IERR(err));
            }
            mustReloadOptions = (setFlags == LIS_SET_FLAG_MUST_RELOAD_OPTIONS);
        }
        if (tolower(name) == string(OPT_NAME_MODE))
        {
            int setFlags = 0;
            union lis_value newValue;
            newValue.string = m_mode.c_str();
            
            if (!LIS_OPT_IS_READABLE(optionDescriptor))
            {
                const string currentValue = value.string;
                if (currentValue != m_mode)
                {
                    err = optionDescriptor->fn.set_value(optionDescriptor, newValue, &setFlags);
                }
            }
            else
            {
                err = optionDescriptor->fn.set_value(optionDescriptor, newValue, &setFlags);
            }
            if (err) // LIS_APPROXIMATE_VALUE, LIS_ACCESS_DENIED, LIS_INVALID_VALUE
            {
                // Don't free options nor return here; use default values.
                if (m_evh)
                    m_evh->OnInsaneError(IERR(err));
            }
            mustReloadOptions = (setFlags == LIS_SET_FLAG_MUST_RELOAD_OPTIONS);
        }
        if (tolower(name) == string(OPT_NAME_RESOLUTION))
        {
            int setFlags = 0;
            union lis_value newValue;
            newValue.integer = m_resolution;
            
            if (!LIS_OPT_IS_READABLE(optionDescriptor))
            {
                const int currentValue = value.integer;
                if (currentValue != m_resolution)
                {
                    err = optionDescriptor->fn.set_value(optionDescriptor, newValue, &setFlags);
                }
            }
            else
            {
                err = optionDescriptor->fn.set_value(optionDescriptor, newValue, &setFlags);
            }
            if (err) // LIS_APPROXIMATE_VALUE, LIS_ACCESS_DENIED, LIS_INVALID_VALUE
            {
                // Don't free options nor return here; use default values.
                if (m_evh)
                    m_evh->OnInsaneError(IERR(err));
            }
            mustReloadOptions = (setFlags == LIS_SET_FLAG_MUST_RELOAD_OPTIONS);
        }
    }
    // Reload if indicated.
    if (mustReloadOptions)
    {
        loadOptions();
        if (err)
        {
            FREE(options);
            if (m_evh)
                m_evh->OnInsaneError(IERR(err));
            return false;
        }
    }
    return true;
}

// -----------------------------------------------------------------------------
InsaneWorkerEvent::InsaneWorkerEvent()
{}

InsaneWorkerEvent::~InsaneWorkerEvent()
{}

void InsaneWorkerEvent::OnInsaneError(const std::string& message)
{}
