/* 
 * File:   XSaneWorker.cpp
 * Author: Saleem EDAH-TALLY - nmset@yandex.com
 * License : GPL
 * Copyright Saleem EDAH-TALLY - © 2017
 * 
 * Created on 5 mars 2017, 11:20
 */

#include "gcc_version.h"
#include "XSaneWorker.h"
#include "MiscTools.h"
#include "globals.h"
#include "DBusCB7Workers.h"
#include "DBusFreeDesktop.h"
#include "PdfBinder.h"
#include <iostream>
#include <vector>
using namespace std;

#define COUNTER_LEN 2

DBus::BusDispatcher dispatcher;

XSaneWorker::XSaneWorker(XCB7 * parent, TimeredStatusBar * sb)
{
    m_owner = parent;
    m_lblNewDoc = m_owner->lblNewDoc;
    m_txtNewDoc = m_owner->txtNewDoc;
    m_btnScan = m_owner->btnScan;
    m_btnRefresh = m_owner->btnRefresh;
    m_cmbScanners = m_owner->cmbScanners;
    m_sb = sb;
    m_config = m_owner->GetConfig();

    m_xsaneProxy = NULL;
    m_ADFProgress = 0;
    m_progress = 0;

    m_btnScan->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &XSaneWorker::SANEDoScan, this);

    DBus::_init_threading();
    DBus::default_dispatcher = &dispatcher;
    DBus::Connection bus = DBus::Connection::SessionBus();
    const wxString name = _T("Classeur.CB7-") + wxVariant((long) wxGetProcessId()).GetString();
    bus.request_name(name.utf8_str().data());
    DBusCB7AdaptorWorker * adaptor_worker = new DBusCB7AdaptorWorker(bus, m_owner);
    adaptor_worker->Run();

    m_cmbScanners->Bind(wxEVT_COMBOBOX, &XSaneWorker::OnScannerChanged, this);
}

XSaneWorker::~XSaneWorker()
{
    delete m_xsaneProxy;
}

bool XSaneWorker::StartXSaneProxy()
{
    static DBus::Connection bus = DBus::Connection::SessionBus();
    // We have checked a device has been selected
    const int selection = m_cmbScanners->GetSelection();
    const XClientData * clientData = static_cast<const XClientData*> (m_cmbScanners->GetClientObject(selection));
    wxVariant serviceName = clientData->GetData();
    try
    {
        if (!m_xsaneProxy) m_xsaneProxy = new XSaneDBusProxy(bus, serviceName.GetString().utf8_str().data(), this);
        if (!m_xsaneProxy->conn().has_name("org.xsane.remote"))
        {
            m_xsaneProxy->IsAlive(); // Generate an error and return;
        }
        return true;
    }
    catch (DBus::Error e)
    {
        wxDELETE(m_xsaneProxy);
        MiscTools::MessageBox(wxString(e.message()) + _NL_ + wxString(e.name()), true);
    }
    return false;
}

wxString XSaneWorker::SANEGetFileName()
{
    static DBus::Connection bus = DBus::Connection::SessionBus();
    try
    {
        XSaneDBusProxy proxy(bus, "org.xsane.remote", this);
        return wxString(proxy.GetFileName());
    }
    catch (DBus::Error e)
    {
        MiscTools::MessageBox(wxString(e.message()) + _NL_ + wxString(e.name()), true);
    }
    return wxEmptyString;
}

void XSaneWorker::SANEDoScan(wxCommandEvent& evt)
{
    if (m_cmbScanners->GetCount() == 0 ||
        m_cmbScanners->GetSelection() == 0)
    {
        MiscTools::MessageBox(_("No device selected."), true);
        evt.Skip();
        return;
    }
    if (m_txtNewDoc->IsEmpty())
    {
        MiscTools::MessageBox(_("Missing destination file name."), true);
        evt.Skip();
        return;
    }
    if (!StartXSaneProxy())
    {
        evt.Skip();
        return;
    }
    if (m_xsaneProxy->IsScanning())
    {
        MiscTools::MessageBox(_("Device busy."), true);
        evt.Skip();
        return;
    }

    m_ADFProgress = 0;
    UpdateUIControls(false);
    try
    {
        m_xsaneProxy->SetFileName(m_txtNewDoc->GetValue().ToStdString());
        m_xsaneProxy->SetCounterLength(COUNTER_LEN);

        bool doubleSided;
        m_config->Read(_T("/Scanner/ADF/DoubleSided"), &doubleSided, false);
        bool ADF;
        m_config->Read(_T("/Scanner/ADF/Automatic"), &ADF, false);
        int total;
        m_config->Read(_T("/Scanner/ADF/Total"), &total, 1);
        if (!ADF)
        {
            m_xsaneProxy->SetADFPagesMax(1);
            m_xsaneProxy->SetCounterStep(1);
            m_xsaneProxy->ResetStatus();
        }
        else
        {
            if (doubleSided && total < 2)
            {
                MiscTools::MessageBox(_("Too few copies requested for double sided scan."), true);
                UpdateUIControls(true);
                return;
            }
            const wxString counter = GetFileCounter();
            const wxVariant vcounter(counter);
            if (doubleSided)
            {
                if (vcounter.GetLong() == 0)
                {
                    m_xsaneProxy->SetADFPagesMax((total / 2) + (total % 2));
                    m_xsaneProxy->SetCounterStep(COUNTER_LEN);
                    m_xsaneProxy->ResetStatus();
                    m_sb->SetTransientText(_("Starting : FRONT face."));
                }
                else
                {
                    if (!UpdateFileCounter())
                    {
                        MiscTools::MessageBox(_("Unexpected error. Can't scan.\nYou may continue one page at a time."));
                        UpdateUIControls(true);
                        return;
                    }
                    m_xsaneProxy->SetCounterStep(-1 * COUNTER_LEN);
                    m_xsaneProxy->SetADFPagesMax((total / 2));
                    m_xsaneProxy->ResetStatus();
                    m_sb->SetTransientText(_("Starting : BACK face."));
                }
            }
            else
            {
                m_xsaneProxy->SetADFPagesMax(total);
                m_xsaneProxy->SetCounterStep(1);
                m_xsaneProxy->ResetStatus();
            }
        }

        m_xsaneProxy->DoScan(); // Is Asynchronous
    }
    catch (DBus::Error e)
    {
        UpdateUIControls(true);
        MiscTools::MessageBox(wxString(e.message()) + _NL_ + wxString(e.name()), true);
    }
}

bool XSaneWorker::IsScanning() const
{
    if (!m_xsaneProxy) return false;
    return m_xsaneProxy->IsScanning();
}

wxString XSaneWorker::GetFileCounter()
{
    const wxString fname(wxString(m_xsaneProxy->GetFileName()));
    if (fname.IsEmpty())
        return wxEmptyString;
    wxFileName f(fname);
    const wxString baseName = f.GetName();
    // We know how we build the file name
    return baseName.Right(COUNTER_LEN);
}

bool XSaneWorker::UpdateFileCounter()
{
    const wxString fname(wxString(m_xsaneProxy->GetFileName()));
    if (fname.IsEmpty())
        return wxEmptyString;

    const wxFileName f(fname);
    wxString baseName = f.GetName();
    const wxString counter = baseName.Right(COUNTER_LEN);
    const wxVariant vcounter(counter);
    if (vcounter.GetLong() == 0)
        return false;
    int total;
    m_config->Read(_T("/Scanner/ADF/Total"), &total, 1);
    int newCounter;
    if (total % 2)
    {
        newCounter = vcounter.GetLong() - 3;
    }
    else
    {
        newCounter = vcounter.GetLong() - 1;
    }
    wxString sNewCounter = wxVariant(newCounter).GetString();
    const int newCounterLength = sNewCounter.length();
    if (newCounterLength < COUNTER_LEN)
    {
        sNewCounter = wxString('0', COUNTER_LEN - newCounterLength) + sNewCounter;
    }
    baseName = baseName.Left(baseName.length() - counter.length()) + sNewCounter;
    m_txtNewDoc->SetValue(f.GetPath(true) + baseName + _DOT_ + f.GetExt());
    m_xsaneProxy->SetFileName(m_txtNewDoc->GetValue().ToStdString());
    return true;
}

void XSaneWorker::UpdateUIControls(bool enable)
{
    m_btnScan->Enable(enable);
    m_txtNewDoc->Enable(enable);
    m_lblNewDoc->Enable(enable);
    m_btnRefresh->Enable(enable);
    m_cmbScanners->Enable(enable);
}

void XSaneWorker::PopulateScannerList()
{
    DBusFreeDesktop dfd;
    vector<string> scanners = dfd.GetRemoteInstances("org.xsane.remote-");
    if (scanners.size() == 0)
        return;
    m_cmbScanners->Clear();
    vector<string>::iterator it;
    static DBus::Connection bus = DBus::Connection::SessionBus();
    m_cmbScanners->Append(_T(""), (wxClientData*) NULL);
    for (it = scanners.begin(); it != scanners.end(); it++)
    {
        const string serviceName(it->data());
        try {
            XSaneDBusProxy proxy(bus, serviceName.data(), NULL);
            m_cmbScanners->Append(proxy.GetDeviceText(), new XClientData(serviceName));
        }
        catch (DBus::Error e)
        {
            MiscTools::MessageBox(wxString(e.message()) + _NL_ + wxString(e.name()), true);
        }
    }
    if (scanners.size() == 1)
        m_cmbScanners->SetSelection(1);
}

void XSaneWorker::Reset()
{
    m_ADFProgress = 0;
    m_progress = 0;
    m_txtNewDoc->Clear();
    m_files.Clear();
    wxDELETE(m_xsaneProxy);
}

void XSaneWorker::OnScannerChanged(wxCommandEvent& evt)
{
    Reset();
    evt.Skip();
}

void XSaneWorker::SetStatusMessage(const wxString& msg)
{
    if (m_sb)
        m_sb->SetTransientText(msg);
}
////////////////////////////////////////////////////////////////////////////////

XSaneDBusProxy::XSaneDBusProxy(DBus::Connection& connection, const char* service, XSaneWorker * parent)
: DBus::ObjectProxy(connection, "/org/xsane/remote", service)
{
    m_owner = parent;
    m_status = -1;
    try
    {
        m_devADFPagesMax = GetADFPagesMax();
        m_devCounterLen = GetCounterLength();
        m_devCounterStep = GetCounterStep();
    }
    catch (DBus::Error e)
    {
        m_devADFPagesMax = 1;
        m_devCounterLen = COUNTER_LEN;
        m_devCounterStep = 1;
    }
}

XSaneDBusProxy::~XSaneDBusProxy()
{
    if (m_owner == NULL)
        return;
    CallAfter(&XSaneDBusProxy::UpdateUIControls, true);
    try
    { // Reset XSane as it was before us.
        SetADFPagesMax(m_devADFPagesMax);
        SetCounterLength(m_devCounterLen);
        SetCounterStep(m_devCounterStep);
    }
    catch (DBus::Error e)
    {

    }
}

void XSaneDBusProxy::ScanDoneStatus(const int32_t& status)
{
    m_status = status;
    if (status == 5)
    {
        CallAfter(&XSaneDBusProxy::UpdateFileName);
        m_owner->m_ADFProgress++;
    }
    if (status > 0 && status != 5) // /usr/include/sane/sane.h
        m_owner->m_ADFProgress = GetADFPagesMax();
    CallAfter(&XSaneDBusProxy::DoFinalize);
}

void XSaneDBusProxy::ScanDialogCompletion(const bool& completed)
{
    /* Happens before ScanDoneStatus. Happens also after !!!
     * Required for instance if overwriting a file is cancelled.
     * ScanDoneStatus is not called in such a case.
     */
    if (!completed)
    {
        m_owner->m_ADFProgress = GetADFPagesMax();
        CallAfter(&XSaneDBusProxy::DoFinalize);
    }
}

void XSaneDBusProxy::UpdateFileName()
{
    m_owner->m_txtNewDoc->SetValue(GetFileName());
}

void XSaneDBusProxy::DoFinalize()
{
    if (GetADFPagesMax() == m_owner->m_ADFProgress)
    {
        CallAfter(&XSaneDBusProxy::UpdateUIControls, true);
        if (GetCounterStep() == -2)
        {
            if (m_status == 5)
            {
                const wxString msg = _("BACK face scan completed.");
                CallAfter(&XSaneDBusProxy::SetStatusMessage, msg);
            }
        }
        if (GetCounterStep() == 2)
        {
            if (m_status == 5)
            {
                const wxString msg =_("FRONT face scan completed.");
                CallAfter(&XSaneDBusProxy::SetStatusMessage, msg);
            }
        }
        
        bool ADF;
        m_owner->m_config->Read(_T("/Scanner/ADF/Automatic"), &ADF, false);
        bool doubleSided;
        m_owner->m_config->Read(_T("/Scanner/ADF/DoubleSided"), &doubleSided, false);
        int total;
        m_owner->m_config->Read(_T("/Scanner/ADF/Total"), &total, 1);
        bool bind;
        m_owner->m_config->Read(_T("/Scanner/Bind"), &bind, true);
        
        /*
         * Merge individual PDF files in the end.
         * Use the last individual file name as the final name.
         */
        auto doBind = [&] ()
        {
            if (!bind)
                return;
            if (DoMerge())
            {
                const wxString msg = _("Bind completed.");
                CallAfter(&XSaneDBusProxy::SetStatusMessage, msg);
            }
        };
        
        if (!ADF) // Double sided is meaningless here.
        {
            m_owner->m_progress++; // Number of passes here == number of pages
            if (m_owner->m_progress == total)
            {
                doBind();
                CallAfter(&XSaneDBusProxy::Reset);
            }
        }
        else
        {
            if (doubleSided) // Mysterious passes here : number of pages for the first project, then only once !
            {
                if (GetCounterStep() == -2 && m_status == 5)
                {
                    doBind();
                    CallAfter(&XSaneDBusProxy::Reset);
                }
            }
            else // Only one pass here.
            {
                doBind();
                CallAfter(&XSaneDBusProxy::Reset);
            }
        }
    }
}

void XSaneDBusProxy::UpdateUIControls(bool enable)
{
    m_owner->UpdateUIControls(enable);
}

void XSaneDBusProxy::OnSaveImageAs(const std::string& savedFileName) {
    m_owner->m_files.Add(savedFileName);
    bool ADF;
    m_owner->m_config->Read(_T("/Scanner/ADF/Automatic"), &ADF, false);
    if (m_owner->m_sb && ADF) {
        int total;
        m_owner->m_config->Read(_T("/Scanner/ADF/Total"), &total, 1);
        const wxString msg = wxVariant((long) m_owner->m_ADFProgress + 1).GetString()
                        + wxString(_T(" / ")) + wxVariant((long) total).GetString();
        /*
         * Here m_owner->m_sb->SetTransientText(msg) generates a message
         * that wxTimer in TimeredStatusBar cannot be called from a non-main thread,
         * though XSaneDBusProxy does not extend wxThread.
         * But here is called asynchronously from remote XSane via DBus. This
         * requires DBusCB7Adaptor, which involves wxThread.
         */
        CallAfter(&XSaneDBusProxy::SetStatusMessage, msg);
    }
}

void XSaneDBusProxy::OnCancel() {
    wxString msgMerge = wxEmptyString;
    bool bind;
    m_owner->m_config->Read(_T("/Scanner/ADF/Automatic"), &bind, true);
    if (bind && m_owner->m_files.GetCount() && DoMerge())
        msgMerge = _("Binding of available pages.");
    m_owner->m_files.Clear();
    const wxString msg = _("Cancelled.") + msgMerge;
    CallAfter(&XSaneDBusProxy::SetStatusMessage, msg);
    m_owner->m_ADFProgress = 0;
    m_status = -1;
    CallAfter(&XSaneDBusProxy::UpdateFileName);
    CallAfter(&XSaneDBusProxy::UpdateUIControls, true);
}

void XSaneDBusProxy::AsyncNotify(const wxString& msg)
{
    MiscTools::MessageBox(msg, true);
}

bool XSaneDBusProxy::DoMerge()
{
    wxString lastFormat;
    m_owner->m_config->Read(_T("/Last/Format"), &lastFormat);
    if (lastFormat.Lower() != _T("pdf"))
        return false;
    if (m_owner->m_files.GetCount() < 2)
        return false;
    PdfBinder pdfb;
    m_owner->m_files.Sort();
    const wxString finalFilePath = m_owner->m_files.Last();
    wxString path, name, ext;
    wxFileName::SplitPath(finalFilePath, &path, &name, &ext);
    const wxString tmpFilePath = path + wxFileName::GetPathSeparator()
    + _T(".") + name + _T(".") + ext;
    bool res = pdfb.Merge(m_owner->m_files, tmpFilePath);
    if (!res)
    {
        CallAfter(&XSaneDBusProxy::AsyncNotify, pdfb.GetWhat());
        if (wxFileExists(tmpFilePath))
            wxRemoveFile(tmpFilePath); // Result not checked
    }
    else
    {
        wxArrayString::iterator it;
        for (it = m_owner->m_files.begin(); it != m_owner->m_files.end(); it++)
        {
            if (wxFileExists(*it))
                wxRemoveFile(*it);
        }
        wxRenameFile(tmpFilePath, finalFilePath, true); // true : in case we can't remove. Then, overwrite will fail too!
    }
    return res;
}

void XSaneDBusProxy::Reset()
{
    m_owner->Reset();
}

void XSaneDBusProxy::SetStatusMessage(const wxString& msg)
{
    m_owner->SetStatusMessage(msg);
}
