/* 
 * File:   XSaneWorker.h
 * Author: Saleem EDAH-TALLY - nmset@yandex.com
 * License : GPL
 * Copyright Saleem EDAH-TALLY - © 2017
 *
 * Created on 5 mars 2017, 11:20
 */

#ifndef XSANEWORKER_H
#define XSANEWORKER_H

#include <wx/wx.h>
#include <wx/config.h>
#include "XCB7.h"
#include "xsane_proxy.hpp"
#include "TimeredStatusBar.h"

class XCB7;
class XSaneDBusProxy;

/*
 * Starts a DBus proxy derived from org::xsane::remote_proxy, the real interface
 * to the remote DBus server (adaptor) in a patched XSane application.
 * Provided file dbus_xsane_03.diff should be used to patch xsane-0.999.
 */
class XSaneWorker
{
public:
    friend class XSaneDBusProxy;

    XSaneWorker(XCB7 * parent, TimeredStatusBar * sb);
    virtual ~XSaneWorker();

    bool IsScanning() const;
    void PopulateScannerList();
private:
    XCB7 * m_owner;
    wxStaticText * m_lblNewDoc;
    wxTextCtrl * m_txtNewDoc;
    wxButton * m_btnScan;
    wxButton * m_btnRefresh;
    wxComboBox * m_cmbScanners;
    TimeredStatusBar * m_sb;
    wxConfig * m_config;
    /*
     * Number of scanned pages that we track if ADF.
     * May not be the total of the project, in case of double sided scanning.
     * ::SetADFPagesMax() tells XSane how many pages to scan, ADF or not.
     * It is 1 if no ADF, and we will have to click the scan button many times if total > 1.
     */
    int m_ADFProgress;
    uint m_progress; // Number of scanned pages that we track if no ADF.
    wxArrayString m_files;

    XSaneDBusProxy * m_xsaneProxy;

    bool StartXSaneProxy();
    /**
     * Gets the file name in XSane's text box.
     * This file name will normally have changed at the end of a successful scan,
     * as we always use a counter suffix.
     * @return 
     */
    wxString SANEGetFileName();
    /**
     * Is asynchronous.
     * Known issue : is synchronus in VirtualBox.
     * Unknown behaviour in QEmu.
     */
    void SANEDoScan(wxCommandEvent& evt);

    /**
     * The counter is changed by xsane on successful scans.
     * We want to know where it's up to.
     * @return 
     */
    wxString GetFileCounter();
    /**
     * Updates xsane's counter when we start scanning back face.
     * @return 
     */
    bool UpdateFileCounter();
    /**
     * Enables or disables controls.
     * @param enable
     */
    void UpdateUIControls(bool enable);
    
    /**
     * Deletes the XSane proxy worker and resets default values
     * whenever a new scanner is selected.
     */
    void Reset();
    
    /**
     * Calls Reset();
     */
    void OnScannerChanged(wxCommandEvent& evt);
    
    /**
     * Calls TimeredStatusBar::SetTransientText : its wxTimer must be started from the main thread.
     */
    void SetStatusMessage(const wxString& msg);
};

/*
 * org::xsane::remote_proxy is the real interface to the remote
 * DBus server (adaptor).
 * 
 * Derives from wxEvtHandler to inherit CallAfter().
 */
class XSaneDBusProxy : public org::xsane::remote_proxy,
public DBus::IntrospectableProxy,
public DBus::ObjectProxy,
public wxEvtHandler
{
public:
    XSaneDBusProxy(DBus::Connection &connection, const char * service, XSaneWorker * parent);
    virtual ~XSaneDBusProxy();

    /**
     * Sent back from the adaptor to indicate scan results.
     * 
     * status : defined in /usr/include/sane/sane.h
     */
    virtual void ScanDoneStatus(const int32_t& status);
    /**
     * Happens before ScanDoneStatus.
     * Required for instance if overwriting a file is cancelled.
     * ScanDoneStatus is not called in such a case.
     * @param completed : true only after successful scan.
     */
    virtual void ScanDialogCompletion(const bool& completed);
    /**
     * Updates the file name in CB7's UI. Use with CallAfter().
     */
    void UpdateFileName();
    /**
     * Re-enables relevant widgets and resets things when back face is completed.
     */
    void DoFinalize();

    void ResetStatus()
    {
        m_status = -1;
    }
    
    /**
     * When XSane has finished scanning a page.
     */
    void OnSaveImageAs(const std::string& savedFileName);
    
    /**
     * Scan is cancelled in XSane.
     */
    void OnCancel();
    
    /**
     * Merge individual PDF files.
     */
    bool DoMerge();
    
    /**
     * Asynchronous messages in CallAfter().
     */
    void AsyncNotify(const wxString& msg);
    
    /**
     * To call XSaneWorker::Reset in CallAfter.
     */
    void Reset();
    
    /**
     * To call XSaneWorker::SetStatusMessage in CallAfter.
     */
    void SetStatusMessage(const wxString& msg);

    /**
     * Use with CallAfter. Calls XSaneWorker::UpdateUIControls .
     * @param enable
     */
    void UpdateUIControls(bool enable);

private:

    XSaneWorker * m_owner;
    uint m_devCounterLen;
    int m_devCounterStep;
    uint m_devADFPagesMax;
    int m_status;
};

#endif /* XSANEWORKER_H */

